"""
Flow Layout

Presents in 2D a 1D list of Widgets

Adapted from https://github.com/PySide/Examples/blob/master/examples/layouts/flowlayout.py
"""
from PySide6.QtCore import Qt, QRect, QSize, QPoint
from PySide6.QtWidgets import QLayout, QSizePolicy


class FlowLayout(QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1, expand_last=(False, False)):
        QLayout.__init__(self, parent)
        self.item_list = []
        self._expand_h = expand_last[0]
        self._expand_v = expand_last[1]
        self._expand_last = expand_last[0] or expand_last[1]

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.item_list.append(item)

    def count(self):
        return len(self.item_list)

    def insertWidget(self, index, widget):
        self.addWidget(widget)
        self.item_list.insert(index, self.item_list.pop(-1))

    def itemAt(self, index):
        if index >= 0 and index < len(self.item_list):
            return self.item_list[index]
        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.item_list):
            return self.item_list.pop(index)

        return None

    def removeAt(self, index):
        item = self.takeAt(index)
        if item is None:
            return None

        item.widget().setParent(None)
        item.widget().deleteLater()
        return True

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.do_layout(QRect(0, 0, width, 0))
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.do_layout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize(2 * self.contentsMargins().top(), 2 * self.contentsMargins().top())
        return size

    def do_layout(self, rect, test_only=True):
        margins = self.contentsMargins()
        start_x = rect.x() + margins.left()
        start_y = rect.y() + margins.top()

        x = start_x
        y = start_y
        line_height = 0

        for item in self.item_list:
            widget = item.widget()

            space_x = self.spacing() + widget.style().layoutSpacing(
                QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal
            )
            space_y = self.spacing() + widget.style().layoutSpacing(
                QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical
            ) + 1

            next_x = x + widget.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = start_x
                y = y + line_height + space_y
                next_x = x + widget.sizeHint().width() + space_x
                line_height = 0

            if item == self.item_list[-1] and self._expand_last:
                width = rect.width() - x - 1 if self._expand_h else item.sizeHint().width()
                height = rect.height() - y - 1 if self._expand_v else item.sizeHint().height()
                size = QSize(width, height)
            else:
                size = item.sizeHint()

            if not test_only:
                item.setGeometry(QRect(QPoint(x, y), size))

            x = next_x

            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()



from typing import List

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QGroupBox, QVBoxLayout, QHBoxLayout, QGridLayout, QSizePolicy


def make_group(title: str, widgets: List[QWidget], orientation = Qt.Vertical, tooltip: str = None, fixed_width: int = None, fixed_height: int = None) -> QGroupBox:
    """
    Creates a QGroupBox with given widgets, title and orientation
    """
    group = QGroupBox(title)
    group.setToolTip(tooltip)
    layout = QVBoxLayout(group) if orientation == Qt.Vertical else QHBoxLayout(group)

    for widget in widgets:
        layout.addWidget(widget)

    if fixed_width is not None:
        group.setFixedWidth(fixed_width)

    if fixed_height is not None:
        group.setFixedHeight(fixed_height)

    return group


def make_group_grid(title: str, widgets: List[List[QWidget]], stretch_last_column=False):
    """
    Creates a QGroupBox with given widgets and title
    Layout is QGridLayout
    """
    group = QGroupBox(title)
    layout = QGridLayout(group)

    for row, widget_row in enumerate(widgets):
        for column, widget in enumerate(widget_row):
            layout.addWidget(widget, row, column)

    if stretch_last_column:
        layout.setColumnStretch(len(widgets[0]) - 1, 1)

    return group



from PySide6.QtWidgets import QWidget, QHBoxLayout, QLabel, QSpinBox, QDoubleSpinBox


class _BaseSpinBox(QWidget):  # FIXME autocompletion ?

    _spinBoxClass = None

    def __init__(self, name="", minimum=0, maximum=0, value=0, single_step=1, on_value_changed=None, parent=None):
        QWidget.__init__(self, parent)

        self.spinbox = self._spinBoxClass()
        layout = QHBoxLayout(self)

        self.spinbox.setSingleStep(single_step)
        self.spinbox.setMinimum(minimum)
        self.spinbox.setMaximum(maximum)
        self.spinbox.setValue(value)

        self.label = QLabel(name)

        if on_value_changed is not None:
            self.spinbox.valueChanged.connect(on_value_changed)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.label)
        layout.addWidget(self.spinbox)
        layout.setStretch(0, 100)

    def __getattr__(self, item):
        return getattr(self.spinbox, item)



class SpinBox(_BaseSpinBox):
    """
    Thin wrapper for a QSpinBox that shows its name
    """
    _spinBoxClass = QSpinBox


class DoubleSpinBox(_BaseSpinBox):
    """
    Thin wrapper for a QDoubleSpinBox that shows its name
    """
    _spinBoxClass = QDoubleSpinBox



from typing import List
from PySide6.QtWidgets import QComboBox


def update(combo: QComboBox, items: List[str], current_text="", reset: bool = False):
    if not reset:
        current_text = combo.currentText() if not current_text else current_text
    combo.blockSignals(True)
    combo.clear()

    if not items:
        return

    combo.addItems(items)
    try:
        combo.setCurrentIndex(items.index(current_text))
    except ValueError:
        combo.setCurrentIndex(-1)

    combo.blockSignals(False)


# def set_current_item(combo: QComboBox, item_text):
#     items = [combo.itemText(i) for i in range(combo.count())]
#     try:
#         combo.setCurrentIndex(items.index(item_text))
#         return True
#     except ValueError:
#         return False



from typing import Callable

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QSlider


class Slider(QWidget):  # FIXME autocompletion ?
    """
    Thin wrapper for a QSlider that shows its value
    """
    def __init__(self, name=None, is_vertical=False, minimum=0, maximum=0, value=0, single_step=1, on_value_changed: Callable=None, parent=None):
        QWidget.__init__(self, parent)

        value = min(max(minimum, value), maximum)

        if is_vertical:
            self.slider = QSlider(Qt.Vertical)
            layout = QVBoxLayout(self)
        else:
            self.slider = QSlider(Qt.Horizontal)
            layout = QHBoxLayout(self)

        self.slider.setSingleStep(single_step)
        self.slider.setMinimum(minimum)
        self.slider.setMaximum(maximum)
        self.slider.setValue(value)

        self.label = QLabel()
        self.label.setFixedWidth(40)  # FIXME compute width from min/max ?

        self.slider.valueChanged.connect(self._update_label)
        if on_value_changed is not None:
            self.slider.valueChanged.connect(on_value_changed)

        layout.setContentsMargins(0, 0, 0, 0)
        if name is not None:
            layout.addWidget(QLabel(name))
        layout.addWidget(self.slider)
        layout.addWidget(self.label)

        self._update_label(value)

    def _update_label(self, value):
        self.label.setText(f"{value}")

    def hasFocus(self):
        return self.slider.hasFocus()

    def __getattr__(self, item):
        return getattr(self.slider, item)



import functools

from PySide6.QtWidgets import QApplication
from PySide6.QtCore import Qt


def hourglass_wrapper(func, parent=None):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with Hourglass(parent):
            return func(*args, **kwargs)

    return wrapper


class Hourglass:
    """
    Context manager to freeze and show a hourglass
    ````
    with Hourglass(self):
        # do lengthy stuff
    ````
    """

    def __init__(self, disable=True, parent=None):
        self._disable = disable
        if parent is None:
            parent = QApplication.topLevelWidgets()[0]  # TODO : are we sure of this ?
        self._parent = parent
        self._widget = QApplication.focusWidget()

    def __enter__(self):
        QApplication.setOverrideCursor(Qt.WaitCursor)
        if self._disable:
            self._parent.setEnabled(False)
        QApplication.processEvents()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if self._disable:
                self._parent.setEnabled(True)
            self._widget.setFocus()
        except AttributeError:
            pass
        finally:
            QApplication.restoreOverrideCursor()
            QApplication.processEvents()



import re

from PySide6.QtCore import Signal, Qt, QEvent
from PySide6.QtWidgets import QApplication, QWidget, QComboBox, QCompleter, QPushButton

from pyside6helpers import icons
from pyside6helpers.flow_layout import FlowLayout


_REGEXP = re.compile(r'<tag:([^>]+)>')


class _Clipboard(object):
    """
    Thin wrapper around QApplication's clipboard
    """

    @staticmethod
    def get():
        clipboard = QApplication.clipboard()
        return clipboard.text()

    @staticmethod
    def set(text_):
        clipboard = QApplication.clipboard()
        clipboard.setText(text_)


class _ButtonSelectionModel(object):
    def __init__(self):
        self._names = list()
        self._index = 0
        self._shift_index = 0
        self._is_shift = False

    @property
    def names(self):
        return self._names

    @names.setter
    def names(self, names):
        self._names = names
        self._index = 0
        self._shift_index = 0

    @property
    def current(self):
        """
        Name of the selected Button (active)
        :return:
        """
        return self.names[self._index]

    def _update_shift(self, is_shift=False):
        """
        Checks whether shift status has changed, then stores the index at which shift was pressed
        :param is_shift: Shift status
        """
        if self._is_shift != is_shift:
            self._shift_index = self._index

        self._is_shift = is_shift

    def desselect(self):
        self._index = 0
        self._shift_index = 0
        self._is_shift = False

    def left(self, is_shift=False):
        """
        Move selection one step left
        :param is_shift: if shift key is pressed
        """
        self._update_shift(is_shift)
        self._index = max(self._index - 1, -len(self.names))

    def right(self, is_shift=False):
        """
        Move selection one step right
        :param is_shift: if shift key is pressed
        """
        self._update_shift(is_shift)
        self._index = min(0, self._index + 1)

    def selected(self):
        """
        Names of the selected Buttons
        :return:
        """
        if self._is_shift:
            if self._index == self._shift_index:
                return [self.names[self._index]]

            elif self._index < self._shift_index:
                self._shift_index = min(-1, self._shift_index)
                indexes = range(self._index, self._shift_index + 1)
                return [self.names[i] for i in indexes]

            else:
                self._index = min(-1, self._index)
                indexes = range(self._shift_index, self._index + 1)
                return [self.names[i] for i in indexes]

        if self._index == 0:
            return list()

        return [self.names[self._index]]


class TagBar(QWidget):
    tags_changed = Signal(list)
    tag_created = Signal(str)

    def __init__(self, parent=None):
        super(TagBar, self).__init__(parent=parent)
        self.setFocusPolicy(Qt.StrongFocus)

        self._model = _ButtonSelectionModel()
        self._buttons = list()

        self._autocompletables = list()

        self._editor = QComboBox()
        self._editor.name = '6ze8f7'
        self._editor.currentIndexChanged.connect(self._index_changed)
        self._editor.setEditable(True)
        self._editor.installEventFilter(self)

        self._layout = FlowLayout(self, expand_last=(True, False))
        self._layout.addWidget(self._editor)

    @property
    def tags(self):
        return [button.name for button in self._buttons]

    @property
    def autocompletables(self):
        return self._autocompletables

    @autocompletables.setter
    def autocompletables(self, items):
        self._autocompletables = items
        completer = QCompleter(self._autocompletables, self)
        completer.setCaseSensitivity(Qt.CaseInsensitive)
        self._editor.blockSignals(True)
        self._editor.setCompleter(completer)
        self._editor.clear()
        self._editor.addItems(items)
        self._editor.blockSignals(False)
        self._editor.setCurrentIndex(-1)

    def _index_changed(self):
        """
        When a choice is made in the combobox
        """
        self.add_button_from_text()

    def _update_model(self):
        """
        Updates model with current button names
        """
        self._model.names = [button.name for button in self._buttons]

    def _new_button(self, tag):
        """
        Creates a new widget button
        :param tag: A valid tag
        :return: the new QPushButton, None otherwise
        """
        if tag in self._model.names:
            return

        if '<' in tag or '>' in tag or not tag:
            return

        button = QPushButton(tag)
        button.setIcon(icons.cancel())
        button.clicked.connect(self._button_clicked)
        button.setFocusPolicy(Qt.NoFocus)
        button.installEventFilter(self)
        button.name = tag

        self._layout.insertWidget(self._layout.count() - 1, button)
        self._buttons.append(button)
        self._update_model()

        self.tags_changed.emit(self.tags)
        if tag not in self._autocompletables:
            self.tag_created.emit(tag)

        return button

    def _button_clicked(self):
        """
        Deletes button when clicked
        """
        self._delete_button(self.sender().name)

    def _set_text_editable(self, is_enabled):
        """
        Defines if text edit is editable
        :param is_enabled: bool
        """
        self._editor.lineEdit().setReadOnly(not is_enabled)

    def _set_text_cursor_position(self, position):
        """
        Defines cursor position of text edit
        :param position: int
        """
        self._editor.lineEdit().setCursorPosition(position)

    def _select_buttons(self):
        """
        Updates buttons states according to model
        """
        _ = [button.setDown(False) for button in self._buttons]

        for name in self._model.selected():
            index = self._model.names.index(name)
            button = self._buttons[index]
            button.setDown(True)

    def _delete_button(self, name):
        """
        Deletes a button by its name
        :param name: Button name
        """
        index = [item.widget().name for item in self._layout.item_list].index(name)
        if self._layout.removeAt(index) is not None:
            self._buttons.pop(index)

        self._update_model()

        self.tags_changed.emit(self.tags)

    def _delete_selected_buttons(self):
        """
        Removes selected buttons
        """
        names = self._model.selected()

        # TODO : this should be dealt with by caller !!
        if not names:
            names = [self._model.names[-1]]

        for name in names:
            self._delete_button(name)

    def desselect(self):
        """
        Clears button selection
        """
        self._model.desselect()
        self._select_buttons()

    def add_button_from_text(self):
        """
        Adds a new button, according to editable text content, clears text editor
        """
        text = self._editor.currentText()
        self._new_button(text)

        self._editor.blockSignals(True)
        self._editor.setEditText('')
        self._editor.blockSignals(False)

    def set_tags(self, tags):
        """
        Clears buttons and sets them
        :param tags: list of str
        """
        for name in self._model.names:
            self._delete_button(name)

        for name in tags:
            self._new_button(name)

    def eventFilter(self, object_, event):
        type_ = event.type()
        text = self._editor.currentText()
        cursor = self._editor.lineEdit().cursorPosition()
        text_selected = self._editor.lineEdit().selectedText()
        is_shift = QApplication.keyboardModifiers() & Qt.ShiftModifier == Qt.ShiftModifier
        is_ctrl = QApplication.keyboardModifiers() & Qt.ControlModifier == Qt.ControlModifier

        if type_ == QEvent.KeyPress:
            key = event.key()

            # TODO : better unification with model._delete_selected_buttons()
            if key == Qt.Key_Backspace and cursor == 0 and self._buttons and not text_selected:
                self._delete_selected_buttons()

            if key == Qt.Key_Delete and self._model.selected():
                self._delete_selected_buttons()

            elif key in (Qt.Key_Enter, Qt.Key_Return) and text:
                self.add_button_from_text()
                return True

            elif key == Qt.Key_Left and cursor == 0 and self._buttons:
                self._model.left(is_shift)
                self._select_buttons()
                return True

            elif key == Qt.Key_Right and cursor == 0:
                self._model.right(is_shift)
                self._select_buttons()
                return bool(self._model.selected())

            elif key == Qt.Key_Down and cursor == 0:
                self._editor.showPopup()
                return True

            elif key == Qt.Key_C and is_ctrl:
                tags = ' '.join(['<tag:{}>'.format(tag) for tag in self._model.selected()])
                _Clipboard.set(tags)
                self.desselect()
                return True

            elif key == Qt.Key_V and is_ctrl:
                for tag in _REGEXP.findall(_Clipboard.get()):
                    self._new_button(tag)
                self.desselect()
                return True

            elif key == Qt.Key_Tab:
                parent = self.parent()
                if parent: parent.focusNextChild()

            elif key == Qt.Key_Backtab:
                parent = self.parent()
                if parent: parent.focusPreviousChild()

        return QWidget.eventFilter(self, object_, event)



from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QKeySequence
from PySide6.QtWidgets import QTableView


def resize_columns_to_content_with_padding(table_view: QTableView, padding: int):
    table_view.resizeColumnsToContents()
    for section in range(table_view.horizontalHeader().count()):
        table_view.setColumnWidth(section, table_view.columnWidth(section) + padding)


class TableView(QTableView):
    beginPaste = Signal()
    endPaste = Signal()
    """
    QTableView with deletion key functionality
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self._internal_clipboard: dict[tuple[int, int], str] = dict()

    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Delete, Qt.Key_Backspace):
            selected_indexes = self.selectionModel().selectedIndexes()

            for index in selected_indexes:
                self.model().setData(index, "", Qt.EditRole)

        elif event.matches(QKeySequence.StandardKey.Copy):
            self._copy_to_clipboard()

        elif event.matches(QKeySequence.StandardKey.Paste):
            self._paste_from_clipboard()

        super().keyPressEvent(event)

    def _copy_to_clipboard(self):
        selected_indexes = self.selectionModel().selectedIndexes()
        if not selected_indexes:
            return

        self._internal_clipboard = dict()
        selected_indexes = sorted(
            selected_indexes, key=lambda index: (index.row(), index.column())
        )
        start_row = selected_indexes[0].row()
        start_col = selected_indexes[0].column()

        for index in selected_indexes:
            self._internal_clipboard[(index.row() - start_row, index.column() - start_col)] = index.data(Qt.DisplayRole) or ""

    def _paste_from_clipboard(self):
        if not self._internal_clipboard:
            return

        selected_indexes = self.selectionModel().selectedIndexes()
        if not selected_indexes:
            return

        selected_indexes = sorted(
            selected_indexes, key=lambda index: (index.row(), index.column())
        )
        start_row = selected_indexes[0].row()
        start_col = selected_indexes[0].column()

        self.beginPaste.emit()
        if len(self._internal_clipboard) == 1:
            for index in selected_indexes:
                self.model().setData(index, self._internal_clipboard[(0, 0)], Qt.EditRole)
        else:
            for (row, column), value in self._internal_clipboard.items():
                if not value:
                    continue
                self.model().setData(self.model().index(start_row + row, start_col + column), value, Qt.EditRole)

        self.endPaste.emit()



from functools import cache
from importlib import resources

from PySide6.QtGui import QIcon, QPixmap, QColor, Qt


def make_path(filepath: str) -> str:
    resource_path = resources.files("pyside6helpers.resources").joinpath(filepath)
    with resources.as_file(resource_path) as p:
        return str(p)


def _make_color(filepath, color: QColor):
    pixmap = QPixmap(filepath)
    mask = pixmap.mask()
    white_pixmap = QPixmap(pixmap.size())
    white_pixmap.fill(color)
    white_pixmap.setMask(mask)
    return white_pixmap


@cache
def align_center(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/align-center.png"), color))


@cache
def align_right(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/align-right.png"), color))


@cache
def back(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/back.png"), color))


@cache
def bar_chart(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/bar-chart.png"), color))


@cache
def barcode(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/barcode.png"), color))


@cache
def book(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/book.png"), color))


@cache
def bookmark(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/bookmark.png"), color))


@cache
def briefcase(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/briefcase.png"), color))


@cache
def calendar(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/calendar.png"), color))


@cache
def cancel(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/cancel.png"), color))


@cache
def cells(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/cells.png"), color))


@cache
def certificate(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/certificate.png"), color))


@cache
def chat(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/chat.png"), color))


@cache
def check(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/check.png"), color))


@cache
def clock(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/clock.png"), color))


@cache
def cloud_computing(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/cloud-computing.png"), color))


@cache
def credit_card(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/credit-card.png"), color))


@cache
def csv_file(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/csv-file.png"), color))


@cache
def dashboard(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/dashboard.png"), color))


@cache
def disabled(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/disabled.png"), color))


@cache
def diskette(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/diskette.png"), color))


@cache
def dislike(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/dislike.png"), color))


@cache
def down_arrow(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/down-arrow.png"), color))


@cache
def download(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/download.png"), color))


@cache
def earth_grid(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/earth-grid.png"), color))


@cache
def eject(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/eject.png"), color))


@cache
def email(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/email.png"), color))


@cache
def equalizer(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/equalizer.png"), color))


@cache
def error(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/error.png"), color))


@cache
def exclamation(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/exclamation.png"), color))


@cache
def file(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/file.png"), color))


@cache
def filter(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/filter.png"), color))


@cache
def flag(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/flag.png"), color))


@cache
def flash(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/flash.png"), color))


@cache
def folder(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/folder.png"), color))


@cache
def headphone(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/headphone.png"), color))


@cache
def heart(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/heart.png"), color))


@cache
def home(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/home.png"), color))


@cache
def image(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/image.png"), color))


@cache
def inbox(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/inbox.png"), color))


@cache
def information(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/information.png"), color))


@cache
def justify(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/justify.png"), color))


@cache
def laptop(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/laptop.png"), color))


@cache
def left_align(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/left-align.png"), color))


@cache
def left_arrow(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/left-arrow.png"), color))


@cache
def levels(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/levels.png"), color))


@cache
def lightbulb(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/lightbulb.png"), color))


@cache
def like(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/like.png"), color))


@cache
def link(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/link.png"), color))


@cache
def list(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/list.png"), color))


@cache
def login(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/login.png"), color))


@cache
def logout(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/logout.png"), color))


@cache
def menu(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/menu.png"), color))


@cache
def minus(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/minus.png"), color))


@cache
def more(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/more.png"), color))


@cache
def move(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/move.png"), color))


@cache
def mute(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/mute.png"), color))


@cache
def next(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/next.png"), color))


@cache
def notification(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/notification.png"), color))


@cache
def padlock(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/padlock.png"), color))


@cache
def pause(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/pause.png"), color))


@cache
def pencil(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/pencil.png"), color))


@cache
def photo_camera(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/photo-camera.png"), color))


@cache
def placeholder(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/placeholder.png"), color))


@cache
def play_button(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/play-button.png"), color))


@cache
def plus(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/plus.png"), color))


@cache
def power_button(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/power-button.png"), color))


@cache
def print(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/print.png"), color))


@cache
def push_pin(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/push-pin.png"), color))


@cache
def qr_code(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/qr-code.png"), color))


@cache
def question_bubble(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/question-bubble.png"), color))


@cache
def question(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/question.png"), color))


@cache
def refresh(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/refresh.png"), color))


@cache
def resize(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/resize.png"), color))


@cache
def right_arrow(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/right-arrow.png"), color))


@cache
def screenshot(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/screenshot.png"), color))


@cache
def search(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/search.png"), color))


@cache
def settings(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/settings.png"), color))


@cache
def share(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/share.png"), color))


@cache
def shield(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/shield.png"), color))


@cache
def shopping_cart(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/shopping-cart.png"), color))


@cache
def shuffle(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/shuffle.png"), color))


@cache
def smartphone(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/smartphone.png"), color))


@cache
def star(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/star.png"), color))


@cache
def stop(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/stop.png"), color))


@cache
def tablet(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/tablet.png"), color))


@cache
def tag(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/tag.png"), color))


@cache
def tasks(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/tasks.png"), color))


@cache
def telephone(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/telephone.png"), color))


@cache
def trash(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/trash.png"), color))


@cache
def unlock(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/unlock.png"), color))


@cache
def up_arrow(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/up-arrow.png"), color))


@cache
def upload(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/upload.png"), color))


@cache
def user(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/user.png"), color))


@cache
def video_camera(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/video-camera.png"), color))


@cache
def vision_stroked(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/vision-stroked.png"), color))


@cache
def vision(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/vision.png"), color))


@cache
def volume(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/volume.png"), color))


@cache
def warning(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/warning.png"), color))


@cache
def wifi(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/wifi.png"), color))


@cache
def zoom_in(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/zoom-in.png"), color))


@cache
def zoom_out(color: QColor = Qt.white) -> QIcon:
    return QIcon(_make_color(make_path("icons/zoom-out.png"), color))



from PySide6.QtWidgets import QFileDialog


def make_save_hook(title, name_filter, working_directory, callback, parent=None):
    """Returns a callable that will getSaveFileName() and calls callback() if a file has been chosen"""
    def save_hook(self):
        dialog = QFileDialog()
        filepath, extension = dialog.getSaveFileName(parent, title, working_directory, name_filter)
        if filepath:
            callback(filepath)

    return save_hook


def make_open_hook(title, name_filter, working_directory, callback, parent=None):
    """Returns a callable that will getOpenFileName() and calls callback() if a file has been chosen"""
    def save_hook(self):
        dialog = QFileDialog()
        filepath, extension = dialog.getOpenFileName(parent, title, working_directory, name_filter)
        if filepath:
            callback(filepath)

    return save_hook



from PySide6.QtWidgets import (QTabWidget, QWidget)


def make_tabs(widgets: dict[str, QWidget]):
    tabs = QTabWidget()
    for caption, widget in widgets.items():
        tabs.addTab(widget, caption)

    return tabs



import sys

from PySide6.QtWidgets import QApplication, QPushButton, QWidget, QVBoxLayout
from PySide6.QtCore import Qt, QPropertyAnimation, Property, Signal, QEasingCurve
from PySide6.QtGui import QPainter, QPalette


class ConfirmButton(QPushButton):

    def __init__(self, text="", confirm_text="Sure ?", duration=1000, parent=None):
        super().__init__(text, parent)
        self._default_text = text
        self._confirm_text = confirm_text
        self._progress = 0.0

        # Setup the animation
        self.animation = QPropertyAnimation(self, b"progress")
        self.animation.setDuration(duration)
        self.animation.setStartValue(0.0)
        self.animation.setEndValue(100.0)
        # Use a linear curve for a steady progress bar
        self.animation.setEasingCurve(QEasingCurve.Type.Linear)

        # Connect animation finished signal
        self.animation.finished.connect(self._on_animation_finished)

    @Property(float)
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value
        self.update()  # Trigger a repaint to show the progress

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.setText(self._confirm_text)
            self.animation.start()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            if self.animation.state() == QPropertyAnimation.State.Running:
                self.animation.stop()
                self._reset_button()
                return

        super().mouseReleaseEvent(event)

    def _on_animation_finished(self):
        # Only fire if we actually reached 100%
        if self._progress >= 100.0:
            self.clicked.emit()

        self._reset_button()

    def _reset_button(self):
        self._progress = 0.0
        self.setText(self._default_text)
        self.update()

    def paintEvent(self, event):
        # Draw the standard button first
        super().paintEvent(event)

        if self._progress > 0:
            painter = QPainter(self)
            # Create a semi-transparent overlay color
            color = self.palette().color(QPalette.ColorRole.Highlight)
            color.setAlpha(100)

            # Calculate the width of the progress bar based on current progress
            width = int(self.width() * (self._progress / 100.0))

            # Draw the progress rectangle
            painter.fillRect(0, 0, width, self.height(), color)
            painter.end()


# Example Usage
if __name__ == "__main__":
    app = QApplication(sys.argv)

    window = QWidget()
    layout = QVBoxLayout(window)

    btn = ConfirmButton("Hold to Delete", "Sure ?", duration=1000)
    btn.setMinimumHeight(50)
    btn.setStyleSheet("font-size: 16px; padding: 10px; background-color: lightgray;")

    # Connect to our custom signal
    btn.clicked.connect(lambda: print("Clicked"))

    layout.addWidget(btn)
    window.show()
    sys.exit(app.exec())



from PySide6.QtCore import QSettings, Signal
from PySide6.QtGui import QPixmap, QCursor, Qt
from PySide6.QtWidgets import QLabel, QMainWindow, QApplication


class MainWindow(QMainWindow):
    firstTimeShown = Signal()
    shown = Signal()

    def __init__(self, logo_filepath: str = "", settings_tuple: tuple[str, str] = tuple()):
        super().__init__()

        if logo_filepath:
            logo_pixmap = QPixmap(logo_filepath)
            logo_label = QLabel()
            logo_label.setPixmap(logo_pixmap)
            logo_label.setScaledContents(True)  # TODO necessary ?
            logo_label.setContentsMargins(2, 2, 5, 0)
            self.statusBar().addPermanentWidget(logo_label)

        organisation = QApplication.instance().organizationName()
        application = QApplication.instance().applicationName()
        self._settings_tuple = (organisation, application) if settings_tuple == tuple() else settings_tuple

        self._fist_time_shown = True

        self.load_geometry()

    def closeEvent(self, event):
        self.save_geometry()
        super().closeEvent(event)
        event.accept()

    def showEvent(self, event):
        self.shown.emit()

        if self._fist_time_shown:
            self._fist_time_shown = False
            self.firstTimeShown.emit()

        event.accept()

    def set_wait(self, is_wait: bool):
        if is_wait:
            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        else:
            QApplication.restoreOverrideCursor()

        self.setEnabled(not is_wait)

    def save_geometry(self):
        if self._settings_tuple:
            settings = QSettings(*self._settings_tuple)
            settings.setValue('geometry', self.saveGeometry())
            settings.setValue('state', self.saveState())

    def load_geometry(self):
        if self._settings_tuple:
            settings = QSettings(*self._settings_tuple)
            self.restoreGeometry(settings.value('geometry'))
            self.restoreState(settings.value('state'))



from PySide6.QtWidgets import QWidget, QVBoxLayout, QSlider, QFrame
from PySide6.QtCore import Qt, Signal, QPointF, QSize
from PySide6.QtGui import QPainter, QColor, QPixmap
import math


class ColorPreview(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumHeight(40)
        self.setFrameShape(QFrame.Box)
        self.setFrameShadow(QFrame.Sunken)
        self._color = QColor(255, 255, 255)

    def setColor(self, color):
        self._color = color
        self.update()

    def paintEvent(self, event):
        super().paintEvent(event)
        painter = QPainter(self)
        painter.fillRect(self.rect().adjusted(1, 1, -1, -1), self._color)


class ColorWheel(QWidget):
    colorChanged = Signal(QColor)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumSize(200, 200)
        self.hue = 0
        self.saturation = 0
        self.value = 255
        self.pressed = False
        self._wheel_pixmap = None
        self._wheel_rect = None
        self._resize_timer = None

    def resizeEvent(self, event):
        super().resizeEvent(event)
        # Invalidate the cached pixmap when resizing
        self._wheel_pixmap = None
        self._wheel_rect = self.rect()

    def drawWheel(self):
        """Create the cached color wheel pixmap"""
        if (self._wheel_pixmap is not None and
                self._wheel_rect == self.rect() and
                self._cached_value == self.value):
            return

        self._wheel_rect = self.rect()
        self._cached_value = self.value
        self._wheel_pixmap = QPixmap(self.size())
        self._wheel_pixmap.fill(Qt.transparent)

        painter = QPainter(self._wheel_pixmap)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Calculate center and radius
        center = QPointF(self.width() / 2, self.height() / 2)
        radius = min(self.width(), self.height()) / 2 - 5

        # Draw color wheel
        for i in range(360):
            painter.setPen(QColor.fromHsv(i, 255, self.value))
            painter.save()
            painter.translate(center)
            painter.rotate(i)
            painter.drawLine(0, 0, radius, 0)
            painter.restore()

        # Draw saturation gradient
        for r in range(int(radius)):
            for angle in range(360):
                x = center.x() + r * math.cos(math.radians(angle))
                y = center.y() + r * math.sin(math.radians(angle))
                sat = (r / radius) * 255
                color = QColor.fromHsv(angle, int(sat), self.value)
                painter.setPen(color)
                painter.drawPoint(int(x), int(y))

        painter.end()

    def paintEvent(self, event):
        self.drawWheel()

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Draw the cached wheel
        painter.drawPixmap(0, 0, self._wheel_pixmap)

        # Draw selector
        if self.pressed:
            center = QPointF(self.width() / 2, self.height() / 2)
            radius = min(self.width(), self.height()) / 2 - 5
            angle = self.hue
            dist = (self.saturation / 255.0) * radius
            x = center.x() + dist * math.cos(math.radians(angle))
            y = center.y() + dist * math.sin(math.radians(angle))
            painter.setPen(Qt.white if self.value < 128 else Qt.black)
            painter.drawEllipse(QPointF(x, y), 4, 4)

    def mousePressEvent(self, event):
        self.pressed = True
        self.updateColor(event.position())
        self.update()

    def mouseMoveEvent(self, event):
        self.updateColor(event.position())
        self.update()

    def mouseReleaseEvent(self, event):
        self.pressed = False
        self.update()

    def updateColor(self, pos):
        center = QPointF(self.width() / 2, self.height() / 2)
        dx = pos.x() - center.x()
        dy = pos.y() - center.y()

        # Calculate hue
        angle = math.degrees(math.atan2(dy, dx))
        if angle < 0:
            angle += 360
        self.hue = angle

        # Calculate saturation
        distance = math.sqrt(dx * dx + dy * dy)
        radius = min(self.width(), self.height()) / 2 - 5
        self.saturation = min(255, int((distance / radius) * 255))

        self.colorChanged.emit(QColor.fromHsv(int(self.hue), self.saturation, self.value))


class ColorPicker(QWidget):
    colorChanged = Signal(QColor)

    def __init__(self, parent=None):
        super().__init__(parent)

        layout = QVBoxLayout(self)

        # Create color wheel
        self.wheel = ColorWheel(self)
        layout.addWidget(self.wheel)

        # Create value (luminance) slider
        self.value_slider = QSlider(Qt.Horizontal)
        self.value_slider.setRange(0, 255)
        self.value_slider.setValue(255)
        layout.addWidget(self.value_slider)

        # Create color preview
        self.preview = ColorPreview(self)
        layout.addWidget(self.preview)

        # Connect signals
        self.wheel.colorChanged.connect(self._onColorChanged)
        self.value_slider.valueChanged.connect(self.onValueChanged)

    def _onColorChanged(self, color):
        self.preview.setColor(color)
        self.colorChanged.emit(color)

    def onValueChanged(self, value):
        self.wheel.value = value
        self.wheel._wheel_pixmap = None  # Invalidate cache when value changes
        self.wheel.update()
        color = QColor.fromHsv(int(self.wheel.hue), self.wheel.saturation, value)
        self._onColorChanged(color)

    def color(self):
        return QColor.fromHsv(int(self.wheel.hue), self.wheel.saturation, self.wheel.value)

    def setColor(self, color):
        h, s, v = color.hue(), color.saturation(), color.value()
        self.wheel.hue = h
        self.wheel.saturation = s
        self.wheel.value = v
        self.value_slider.setValue(v)
        self.preview.setColor(color)
        self.wheel.update()


if __name__ == '__main__':
    from PySide6.QtWidgets import QApplication

    app = QApplication([])
    picker = ColorPicker()
    picker.show()
    app.exec()



import logging
import os.path
import sys

from pythonhelpers.singleton_metaclass import SingletonMetaclass

logging.getLogger(__name__).warning("The pyside6helpers.resources module is deprecated. Use importlib.resources instead.")


class _Resources(metaclass=SingletonMetaclass):
    default_root: str = os.path.join(os.path.abspath(os.path.dirname(__file__)), "resources")
    root: str = ""


def make_path(resource_name: str = "") -> str:
    root = _Resources().default_root if not _Resources().root else _Resources().root
    if not resource_name:
        return root

    return os.path.join(root, resource_name)


def set_root(root_path: str) -> None:
    _Resources().root = root_path


def find(resource_name: str, current_file: str | None = None) -> str:
    if current_file is None:
        current_file = sys.argv[0]

    current_dir = os.path.abspath(os.path.dirname(current_file))
    resources_dir = "resources"

    while current_dir:
        potential_resources_path = os.path.join(current_dir, resources_dir)
        if os.path.isdir(potential_resources_path):
            path = os.path.join(potential_resources_path, resource_name)
            if not os.path.isfile(path):
                raise FileNotFoundError(f"Could not find {resource_name} in {potential_resources_path}")
            return path

        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:
            break

        current_dir = parent_dir

    raise FileNotFoundError(f"Could not find a 'resources' folder starting from {current_file}")





def clear(layout):
    if layout is not None:
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            else:
                clear(item.layout())



from PySide6.QtWidgets import QWidget, QGridLayout


def wrap(widget: QWidget):
    """
    Wraps given widget in a widget with a QGridLayout
    """
    central_widget = QWidget()
    central_layout = QGridLayout(central_widget)
    central_layout.addWidget(widget)
    return central_widget



from PySide6.QtWidgets import QApplication, QMessageBox


def critical_box(message_lines: list[str]) -> None:
    """Show a critical QMessageBox with Ok button"""
    _exec_box(
        icon=QMessageBox.Icon.Critical,
        title="Critical",
        message_lines=message_lines
    )


def warning_box(message_lines: list[str]) -> None:
    """Show a warning QMessageBox with Ok button"""
    _exec_box(
        icon=QMessageBox.Icon.Warning,
        title="Information",
        message_lines=message_lines
    )


def information_box(message_lines: list[str]) -> None:
    """Show an information QMessageBox with Ok button"""
    _exec_box(
        icon=QMessageBox.Icon.Information,
        title="Information",
        message_lines=message_lines
    )


def confirmation_box(message_lines: list[str]) -> bool:
    """Show a confirmation QMessageBox with Ok & Cancel buttons"""
    return _exec_box(
        icon=QMessageBox.Icon.Question,
        title="Are you sure ?",
        message_lines=message_lines,
        has_cancel=True
    )


def _exec_box(icon: QMessageBox.Icon, title: str, message_lines: list[str], has_cancel=False) -> bool:
    _box = QMessageBox(QApplication.activeWindow())
    _box.setIcon(icon)
    _box.setWindowTitle(title)
    _box.setText("\n".join(message_lines))
    if has_cancel:
        _box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
    else:
        _box.setStandardButtons(QMessageBox.Ok)

    result = _box.exec() == QMessageBox.Ok
    _box.deleteLater()

    return result
